from typing import Optional

from ts2vg.graph._natural import _compute_graph
from ts2vg.graph.base import BaseVG


class NaturalVG(BaseVG):
    """
    Representation of a natural visibility graph as obtained from an input time series according
    to the mathematical definitions in:

    Lucas Lacasa et al., "*From time series to complex networks: The visibility graph*", 2008.
    """

    def __init__(self, *, directed: Optional[str] = None, weighted: Optional[str] = None):
        r"""
        Parameters
        ----------
        directed : str, None
            If ``None`` make an undirected graph, otherwise, a directed graph with one of the following
            options:

            ``left_to_right`` :
                Edge directions go from left to right according to the series temporal *x* axis.

            ``top_to_bottom`` :
                Edge directions go from top to bottom according to the series *y* axis.

                .. note::
                    If both endpoints of an edge have the same *y* value then the direction 
                    is ambiguous and no consistent direction is guaranteed.

            Default ``None``.

        weighted : str, None
            If ``None`` make an unweighted graph, otherwise, a weighted graph with one of the following
            options (edges going from point :math:`(p_x, p_y)` to point :math:`(q_x, q_y)`):

            ``distance`` :
                Euclidean distance.
                Calculated as:

                .. math::
                    \sqrt{(q_x - p_x)^2 \cdot (q_y - p_y)^2}

            ``sq_distance`` :
                Squared Euclidean distance (quicker to compute than regular Euclidean distance).
                Calculated as:

                .. math::
                    (q_x - p_x)^2 \cdot (q_y - p_y)^2

            ``v_distance`` :
                Vertical distance.
                Calculated as:

                .. math::
                    q_y - p_y

            ``abs_v_distance`` :
                Absolute vertical distance.
                Calculated as:

                .. math::
                    \left| q_y - p_y \right|

            ``h_distance``:
                Horizontal distance.
                Calculated as:

                .. math::
                    q_x - p_x

            ``abs_h_distance``:
                Absolute horizontal distance.
                Calculated as:

                .. math::
                    \left| q_x - p_x \right|

            ``slope`` :
                Slope, in the range (-∞, +∞).
                Calculated as:

                .. math::
                    \frac{q_y - p_y}{q_x - p_x}

            ``abs_slope`` :
                Absolute slope, in the range (0, +∞).
                Calculated as:

                .. math::
                    \left| \frac{q_y - p_y}{q_x - p_x} \right|

            ``angle`` :
                Slope angle in radians, in the range (-π/2, +π/2).
                Calculated as:

                .. math::
                    \arctan \left( \frac{q_y - p_y}{q_x - p_x} \right)

            ``abs_angle`` :
                Absolute slope angle in radians, in the range (0, +π/2).
                Calculated as:

                .. math::
                    \arctan \left( \left| \frac{q_y - p_y}{q_x - p_x} \right| \right)
            
            Default ``None``.

            .. note::
                Asymmetrical weight functions (like ``v_distance``, ``h_distance``, ``slope``,  ``angle``) depend on the edge direction.
                If the graph is undirected, a 'left to right' edge direction is used by default when computing the weights.
        """
        super().__init__(directed=directed, weighted=weighted)

    def _compute_graph(self, only_degrees: bool):
        return _compute_graph(self.ts, self.xs, self._directed, self._weighted, only_degrees)

    def summary(self):
        self._validate_is_built()

        txt = f"Natural visibility graph"
        if self.is_directed and self.is_weighted:
            txt += " (directed, weighted)"
        elif self.is_directed:
            txt += " (directed)"
        elif self.is_weighted:
            txt += " (weighted)"
        
        txt += f" with {self.n_vertices} vertices and {self.n_edges} edges."

        return txt



import numpy as np
from typing import Optional

_DIRECTED_OPTIONS = {
    None: 0,
    'left_to_right': 1,
    'top_to_bottom': 2,
}

_WEIGHTED_OPTIONS = {
    None: 0,
    'distance': 1,
    'sq_distance': 2,
    'v_distance': 3,
    'abs_v_distance': 4,
    'h_distance': 5,
    'abs_h_distance': 6,
    'slope': 7,
    'abs_slope': 8,
    'angle': 9,
    'abs_angle': 10,
}


class NotBuiltError(Exception):
    """
    Exception class to raise if certain graph attributes or methods are accessed before
    the graph has been built.
    """


class BaseVG:
    """
    Abstract class for a visibility graph (VG).
    
    .. caution::
        Should not be used directly, use one of the subclasses instead,
        e.g :class:`ts2vg.NaturalVG` or :class:`ts2vg.HorizontalVG`.

    Args:
        ts (list of values, or 1d numpy array): Time series data to use as input for the visibility graph.
    """
    def __init__(self, *, directed: Optional[str] = None, weighted: Optional[str] = None):
        self.ts = None
        """1D array of the time series. ``None`` if the graph has not been built yet."""
    
        self.xs = None
        """1D array of the X coordinates of the time series. ``None`` if the graph has not been built yet."""

        self._m = None
        self._edges = None
        self._degrees = None
        self._degrees_in = None
        self._degrees_out = None

        if directed not in _DIRECTED_OPTIONS:
            raise ValueError(f"Invalid 'directed' parameter: {directed}. Must be one of {list(_DIRECTED_OPTIONS.keys())}")

        self.directed = directed
        """`str` indicating the strategy used for the edge directions (same as passed to the constructor). ``None`` if the graph is undirected."""
        self._directed = _DIRECTED_OPTIONS[directed]

        if weighted not in _WEIGHTED_OPTIONS:
            raise ValueError(f"Invalid 'weighted' parameter: {weighted}. Must be one of {list(_WEIGHTED_OPTIONS.keys())}.")

        self.weighted = weighted
        """`str` indicating the strategy used for the edge weights (same as passed to the constructor). ``None`` if the graph is unweighted."""
        self._weighted = _WEIGHTED_OPTIONS[weighted]

    def _validate_is_built(self):
        if self._edges is None:
            raise NotBuiltError("Cannot access graph edges, use 'build' first.")

    def build(self, ts, xs=None, only_degrees: bool = False):
        """
        Compute and build the visibility graph for the given time series `ts`.

        Parameters
        ----------
        ts : 1D array like
            Input time series.

        xs : 1D array like, optional
            X coordinates for the time series.
            Length of ``xs`` must match length of ``ts``.

            If not provided, ``[0, 1, 2...]`` will be used.
        
        only_degrees : bool
            If ``True`` only compute the graph degrees, otherwise compute the whole graph.
            Default ``False``.

        Returns
        -------
            self
        """
        self.ts = np.asarray(ts, dtype=np.float64)

        if self.ts.ndim != 1:
            raise ValueError("Input time series must be one-dimensional.")

        if xs is None:
            self.xs = np.arange(len(ts), dtype=np.float64)
        else:
            if len(xs) != len(self.ts):
                raise ValueError(f"Length of 'xs' ({len(xs)}) does not match length of 'ts' ({len(self.ts)}).")

            self.xs = np.asarray(xs, dtype=np.float64)

            if self.xs.ndim != 1:
                raise ValueError("Input 'xs' series must be one-dimensional.")

            if np.any(np.diff(self.xs) <= 0):
                raise ValueError("Input 'xs' series must be monotonically increasing.")

        if only_degrees and self.is_weighted:
            raise ValueError("Building with 'only_degrees' is only supported for unweighted graphs.")

        self._edges, self._degrees_in, self._degrees_out = self._compute_graph(only_degrees)
        self._degrees = self._degrees_in + self._degrees_out
        
        if only_degrees:  # `_compute_graph` doesn't return valid edges when only_degrees=True
            self._edges = None

        return self

    @property
    def is_directed(self) -> bool:
        """``True`` if the graph is directed, ``False`` otherwise."""
        return self.directed is not None

    @property
    def is_weighted(self) -> bool:
        """``True`` if the graph is weighted, ``False`` otherwise."""
        return self.weighted is not None

    @property
    def n_vertices(self):
        """
        Number of vertices (nodes) in the graph.
        """
        return self.ts.size

    @property
    def n_edges(self):
        """
        Number of edges (links) in the graph.
        """
        if self._m is None:
            if self._edges is not None:
                self._m = len(self._edges)
            elif self._degrees is not None:
                self._m = np.sum(self._degrees, dtype=int) // 2
            else:
                raise NotBuiltError("Cannot access graph edges, use 'build' first.")

        return self._m

    @property
    def edges(self):
        """
        List of edges of the graph.

        Returns the graph edges as an iterable of pairs of integers,
        where each integer corresponds to a node id (assigned sequentially in the same order as the input time series).

        If the graph is weighted, a third value is included for each edge corresponding to its weight.
        """
        self._validate_is_built()

        return self._edges

    @property
    def _edges_array(self):
        arr = np.asarray(self._edges, dtype='int64')  # could be 'uint64' but then it breaks np.bincount

        if self.is_weighted:
            return arr[:, :2]

        return arr

    @property
    def weights(self):
        """
        1D array containing the weights of the edges in the graph (listed in the same order as in :attr:`edges`).
        ``None`` if the graph is unweighted.
        """
        self._validate_is_built()
        
        if self.weighted is None:
            return None
        
        return np.fromiter((w for (_, _, w) in self.edges), dtype='float64', count=self.n_edges)

    @property
    def degrees(self):
        """
        Degree sequence of the graph.

        Returns a list of degree values for each node in the graph, in the same order as the input time series.
        """
        if self._degrees is not None:
            pass
        elif self._edges is not None:
            self._degrees = np.bincount(self._edges_array.flat)
        else:
            raise NotBuiltError("Cannot access graph edges, use 'build' first.")

        return self._degrees

    @property
    def degrees_in(self):
        return self._degrees_in

    @property
    def degrees_out(self):
        return self._degrees_out

    @property
    def degree_counts(self):
        """
        Degree counts of the graph.

        Two lists `ks`, `cs` are returned.
        `cs[i]` is the number of nodes in the graph that have degree `ks[i]`.

        The count of any other degree value not listed in `ks` is 0.
        """
        ks, counts = np.unique(self.degrees, return_counts=True)
        cs = counts

        return ks, cs

    @property
    def degree_distribution(self):
        """
        Degree distribution of the graph.

        Two lists `ks`, `ps` are returned.
        `ps[i]` is the empirical probability that a node in the graph has degree `ks[i]`.

        The probability for any other degree value not listed in `ks` is 0.
        """
        ks, counts = self.degree_counts
        ps = counts / self.n_vertices

        return ks, ps

    def adjacency_matrix(self, triangle='both', use_weights=False, no_weight_value=np.nan):
        """
        Returns the adjacency matrix of the graph.

        Parameters
        ----------
        triangle : str
            One of ``'lower'`` (uses the lower triangle of the matrix),
            ``'upper'`` (uses the upper triangle of the matrix)
            or ``'both'`` (uses both).
            Only applicable for undirected graphs.

            Default ``'both'``.

        use_weights : bool
            If ``True``, return an adjacency matrix containing the edge weights,
            otherwise return a binary adjacency matrix.
            Only applicable for weighted graphs.

            Default ``False``.

        no_weight_value : float 
            The default value used in the matrix for the cases where the nodes are not connected.
            Only applicable for weighted graphs and when using ``use_weights=True``.
            
            Default ``np.nan``.

        Returns
        -------
        2D array
            Adjacency matrix of the graph.

        """
        if triangle not in ['lower', 'upper', 'both']:
            raise ValueError(f"'triangle' must be one of 'lower', 'upper', 'both'. Got '{triangle}'.")

        if use_weights and not self.is_weighted:
            raise ValueError(f"'use_weights=True' only valid for weighted graphs.")

        e = self._edges_array
        w = self.weights

        if self.is_weighted and use_weights:
            m = np.full((self.n_vertices, self.n_vertices), fill_value=no_weight_value, dtype='float64')
            if self.is_directed:
                m[e[:, 0], e[:, 1]] = w
            else:
                if triangle == 'both' or triangle == 'upper':
                    m[e[:, 0], e[:, 1]] = w
                
                if triangle == 'both' or triangle == 'lower':
                    m[e[:, 1], e[:, 0]] = w
        else:
            m = np.zeros((self.n_vertices, self.n_vertices), dtype='uint8')
            if self.is_directed:
                m[e[:, 0], e[:, 1]] = 1
            else:
                if triangle == 'both' or triangle == 'upper':
                    m[e[:, 0], e[:, 1]] = 1
                
                if triangle == 'both' or triangle == 'lower':
                    m[e[:, 1], e[:, 0]] = 1
        
        return m

    def as_igraph(self):
        """
        Returns an `igraph <https://igraph.org/python/>`_ graph object corresponding to this graph.
        The ``igraph`` package is required.
        """
        self._validate_is_built()

        from igraph import Graph

        g = Graph.TupleList(
            self.edges,
            edge_attrs='weight' if self.is_weighted else None,
            directed=self.is_directed
        )

        return g

    def as_networkx(self):
        """
        Returns a `NetworkX <https://networkx.github.io/>`_ graph object corresponding to this graph.
        The ``networkx`` package is required.
        """
        self._validate_is_built()

        from networkx import DiGraph, Graph

        if self.is_directed:
            g = DiGraph()
        else:
            g = Graph()

        if self.is_weighted:
            g.add_weighted_edges_from(self.edges)
        else:
            g.add_edges_from(self.edges)

        return g

    def as_snap(self):
        """
        Returns a `SNAP <https://snap.stanford.edu/snappy/>`_ graph object corresponding to this graph.
        The ``snap`` package is required.
        """
        self._validate_is_built()

        from snap import TUNGraph, TNGraph

        if self.is_weighted:
            raise ValueError("SNAP weighted graphs not currently supported.")

        if self.is_directed:
            g = TNGraph.New(self.n_vertices, self.n_edges)
        else:
            g = TUNGraph.New(self.n_vertices, self.n_edges)

        for i in range(self.n_vertices):
            g.AddNode(i)

        for e in self.edges:
            g.AddEdge(*e)

        return g

    def node_positions(self):
        """
        A dictionary with nodes as keys and positions *(x, y)* as values.
        """

        return {i: (self.xs[i], self.ts[i]) for i in range(self.n_vertices)}

    def summary(self):
        """
        Short text summary of information for the graph.

        Returns
        -------
        str
            A string containing the short summary.
        """
        raise NotImplementedError



import argparse
from pathlib import Path

import numpy as np

from ts2vg import NaturalVG, HorizontalVG
from ts2vg.graph.base import _DIRECTED_OPTIONS, _WEIGHTED_OPTIONS

_OUTPUT_MODES = {
    'el': 'edge list',
    'ds': 'degree sequence',
    'dd': 'degree distribution',
    'dc': 'degree counts'
}

_GRAPH_TYPES = {
    'natural': NaturalVG,
    'horizontal': HorizontalVG,
}


class SmartFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()  
        return argparse.HelpFormatter._split_lines(self, text, width)


def main():
    parser = argparse.ArgumentParser(formatter_class=SmartFormatter, description="Compute the visibility graph from an input time series.")
    parser.add_argument('input', help="Path to the file containing the input time series. Must be a text file with one value per line.")
    parser.add_argument('-o', '--output', help="Path to the file where the output corresponding to the visibility graph will be saved. If not provided, output will go to stdout.")
    parser.add_argument('-t', '--type', choices=_GRAPH_TYPES.keys(), default='natural', help="Type of graph.")
    parser.add_argument('-d', '--directed', choices=[c for c in _DIRECTED_OPTIONS.keys() if c is not None], default=None)
    parser.add_argument('-w', '--weighted', choices=[c for c in _WEIGHTED_OPTIONS.keys() if c is not None], default=None)
    parser.add_argument('-m', '--outputmode', choices=_OUTPUT_MODES.keys(), default='el',  # metavar="",
                        help="R|Graph properties and representation to use for the output. One of:"
                            "\n el : (default) Edge list. Nodes are labelled in the range [0, n-1] in the same order as the input time series."
                            "\n ds :           Degree sequence. Degree values for the nodes in the range [0, n-1] in the same order as the input time series."
                            "\n dd :           Degree distribution. 1st column is a degree value (k) and 2nd column is the empirical probability for that degree k."
                            "\n dc :           egree counts. 1st column is a degree value (k) and 2nd column is the number of nodes with that degree k.")

    args = parser.parse_args()
    input_path = args.input
    output_path = args.output
    gtype = args.type
    directed = args.directed
    weighted = args.weighted
    output_mode = args.outputmode

    output_f = None
    if output_path is not None:
        output_path_ = Path(output_path)
        if not output_path_.parent.exists():
            print(f"ERROR: Output folder for '{output_path}' not found.")
            return
        
        output_f = open(output_path, 'w')

    input_path_ = Path(input_path)
    if not input_path_.is_file():
        print(f"ERROR: Input file '{input_path}' not found.")
        return

    ts = np.loadtxt(input_path_, dtype='float64')

    g = _GRAPH_TYPES[gtype](directed=directed, weighted=weighted)
    g.build(ts)

    if output_mode == 'el':
        es = g.edges
        for (a, b, *w) in es:
            print(a, b, *w, file=output_f)
    elif output_mode == 'ds':
        ds = g.degree_sequence
        for d in ds:
            print(d, file=output_f)
    elif output_mode == 'dd':
        ks, pks = g.degree_distribution
        for k, pk in zip(ks, pks):
            print(k, pk, file=output_f)
    elif output_mode == 'dc':
        ks, nks = g.degree_counts
        for k, nk in zip(ks, nks):
            print(k, nk, file=output_f)

    if output_path is not None:
        output_f.close()
        print(g.summary())
        print(f"Saved {_OUTPUT_MODES[output_mode]} to file: {output_path}")



import sys
from setuptools import setup, find_packages, Extension
from numpy import get_include as get_np_include
from Cython.Build import cythonize


def main():
    include_dirs = [get_np_include()]
    define_macros = [('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]

    extensions = [
        Extension('ts2vg.utils.pairqueue',
                  [f'ts2vg/utils/pairqueue.pyx']),

        Extension('ts2vg.graph._base',
                  [f'ts2vg/graph/_base.pyx'],
                  include_dirs=include_dirs,
                  define_macros=define_macros),

        Extension('ts2vg.graph._natural',
                  [f'ts2vg/graph/_natural.pyx'],
                  include_dirs=include_dirs,
                  define_macros=define_macros),

        Extension('ts2vg.graph._horizontal',
                  [f'ts2vg/graph/_horizontal.pyx'],
                  include_dirs=include_dirs,
                  define_macros=define_macros),
    ]
    
    extensions = cythonize(extensions)

    setup(ext_modules = extensions)

if __name__ == "__main__":
    main()
